/*****************************************************************************/
/*                        PTZ Camera Scheduler                               */
/*                             (C) 2009                                      */
/*                 Adrian Ilie - adyilie@cs.unc.edu                          */
/*                   Computer Science Department                             */
/*            University of North Carolina at Chapel Hill                    */
/*                 http://www.cs.unc.edu/~adyilie                            */
/*                                                                           */
/*****************************************************************************/
/*****************************************************************************/
/* Copyright The University of North Carolina at Chapel Hill.                */
/* All Rights Reserved.                                                      */
/* Permission to use, copy, and modify this software and its documentation   */
/* for educational and research purposes without fee is hereby granted,      */
/* provided that the above copyright notice and the following paragraphs     */
/* appear in all copies.                                                     */
/*                                                                           */
/* IN NO EVENT SHALL THE UNIVERSITY OF NORTH CAROLINA AT CHAPEL HILL BE      */
/* LIABLE TO ANY PARTY FOR DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR         */
/* CONSEQUENTIAL DAMAGES, INCLUDING LOST PROFITS, ARISING OUT OF THE         */
/* USE OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN IF THE UNIVERSITY        */
/* OF NORTH CAROLINA HAVE BEEN ADVISED OF THE POSSIBILITY OF SUCH            */
/* DAMAGES.                                                                  */
/*                                                                           */
/* THE UNIVERSITY OF NORTH CAROLINA SPECIFICALLY DISCLAIMS ANY               */
/* WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF      */
/* MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.  THE SOFTWARE       */
/* PROVIDED HEREUNDER IS ON AN "AS IS" BASIS, AND THE UNIVERSITY OF          */
/* NORTH CAROLINA HAS NO OBLIGATIONS TO PROVIDE MAINTENANCE, SUPPORT,        */
/* UPDATES, ENHANCEMENTS, OR MODIFICATIONS.                                  */
/*                                                                           */
/* If you have any questions regarding this agreement or the licensing of    */
/* this technology, please send them to adyilie@cs.unc.edu                   */
/*****************************************************************************/

#include "EntCon.h"

//Note: Python interface was designed to work with telnet
//so it expects CRLF(\r\n) at the end of each command
//and sends back replies ending in CRLF.
//Entity 0 will manipulate the player in single player games.
//Version 1.0, first public release (June 23, 2009)

bool CEntCon::winSockStarted = false;

CEntCon::CEntCon(const char *pHost, int pPort)
{
	closed = false;
	strcpy(host,pHost);
	port=pPort;
	targetIds=NULL;
	numTargets=0;
	Connect();
}

CEntCon::~CEntCon()
{
	if (!closed) Disconnect();
	if (winSockStarted) WSACleanup( );
	if (targetIds) delete []targetIds;
}

void CEntCon::StartWINSock()
{
	if (winSockStarted) return;
	WORD wVersionRequested;
	WSADATA wsaData;
	int err;
	wVersionRequested = MAKEWORD( 2, 0 );
	err = WSAStartup( wVersionRequested, &wsaData );
	if ( err != 0 )
	{
		Error();
		exit(-1);
		return;
	}
	if ( LOBYTE( wsaData.wVersion ) != 2 ||
		HIBYTE( wsaData.wVersion ) != 0 )
	{//need 2.0
		printf("Need WinSock 2.0\n");
		WSACleanup( );
		exit(-1);
		return; 
	}
	winSockStarted = true;
}

int CEntCon::Connect()
{
	StartWINSock();
	HOSTENT * h;
	bool vb=true;
	int i=0;
	while (vb && host[i])
	{
		vb &= (host[i]=='.')||(isdigit(host[i]));
		i++;
	}
	if (vb)
	{//it's an IP address
		long ipAddr = inet_addr( host );
		h = gethostbyaddr( (char *)&ipAddr, sizeof(long), AF_INET );
		if (h != NULL)
		{
			strcpy(host, h->h_name);
			printf("Resolved hostname to be %s \n", host );
		}
		else Error();
	}
	else
	{//it's a host name
		h = gethostbyname(host);
		if (h == NULL)
		{
			printf("gethostbyname failed with %s \n", host );
			Error();
		}
	}
	if (!h) return -2;
	memset( &addr, 0, sizeof(addr) );
	addr.sin_family = AF_INET;
	addr.sin_port =  htons(port);
	printf("Sending to portNb = %i \n", port);
	memcpy( &addr.sin_addr.s_addr, h->h_addr_list[0], sizeof( addr.sin_addr.s_addr ));

	if ((socketfd = socket(AF_INET, SOCK_STREAM, 0 )) == INVALID_SOCKET)
	{
		printf("Create ");Error();
		return -1;
	}

	int result = connect(socketfd, (sockaddr *) &addr, sizeof(addr) );
	if (result < 0)
	{
		printf("Create ");Error();
	}
	return result;
}

int CEntCon::Send(const char *buf, int size)
{
	if (size==0) size=(int)strlen(buf);
	char *buf2=new char[size+3];
	memset(buf2,0,size+3);
	strncpy(buf2,buf,size);
	if (NewLinePos(buf2)<0)
	{
		buf2[size]='\r';
		size++;
		buf2[size]='\n';
		size++;
		buf2[size]=0;
	}
	int ret = send(socketfd, buf2, size, 0);
	if (ret==SOCKET_ERROR)
	{
		printf("Send ");
		Error();
		delete buf2;
		return -1;
	}
	delete buf2;
	return ret;
}

int CEntCon::Receive(char *buf, int size, int timeout)
{
	if (Probe(timeout))
	{
		int total=0;
		do
		{
			int count=recv( socketfd, buf+total, size-total, 0);
			if (count==SOCKET_ERROR)
			{
				printf("Recv ");
				Error();
				return -1;
			}
			else total+=count;
			int p=NewLinePos(buf,total);
			if (p>0)
			{
				if (p<size) buf[p+1]=0;
				return total;
			}
			if (total<size) Sleep(0);//coming back around, sleep for a bit
		}
		while (total<size);
		return total;
	}
	else
	{
		printf("Recv time out\n");
		return 0;
	}
}

bool CEntCon::Probe(int timeout)
{

	fd_set fd;
	struct timeval t;
	FD_ZERO(&fd);
	FD_SET(socketfd, &fd);
	t.tv_sec = timeout/1000;
	t.tv_usec = (timeout%1000)*1000;
	int n = select((int)socketfd + 1, &fd, 0, 0, &t);
	if (n > 0) return (FD_ISSET(socketfd, &fd)!=0);
	else return false;
}

void CEntCon::Disconnect()
{
	if (socketfd>=0)
	{
		shutdown(socketfd, 2);
		closesocket( socketfd );
	}
	closed=true;
};

int CEntCon::Error()
{
	int err=WSAGetLastError();
	switch (err)
	{
	case WSAEINTR: printf("Interrupted function call.\n");break;
	case WSAEACCES: printf("Permission denied.\n");break;
	case WSAEFAULT: printf("Bad address.\n");break;
	case WSAEINVAL: printf("Invalid argument.\n");break;
	case WSAEMFILE: printf("Too many open files.\n");break;
	case WSAEWOULDBLOCK: printf("Resource temporarily unavailable.\n");break;
	case WSAEINPROGRESS: printf("Operation now in progress.\n");break;
	case WSAEALREADY: printf("Operation already in progress.\n");break;
	case WSAENOTSOCK: printf("Socket operation on nonsocket.\n");break;
	case WSAEDESTADDRREQ: printf("Destination address required.\n");break;
	case WSAEMSGSIZE: printf("Message too long.\n");break;
	case WSAEPROTOTYPE: printf("Protocol wrong type for socket.\n");break;
	case WSAENOPROTOOPT: printf("Bad protocol option.\n");break;
	case WSAEPROTONOSUPPORT: printf("Protocol not supported.\n");break;
	case WSAESOCKTNOSUPPORT: printf("Socket type not supported.\n");break;
	case WSAEOPNOTSUPP: printf("Operation not supported.\n");break;
	case WSAEPFNOSUPPORT: printf("Protocol family not supported.\n");break;
	case WSAEAFNOSUPPORT: printf("Address family not supported.\n");break;
	case WSAEADDRINUSE: printf("Address already in use.\n");break;
	case WSAEADDRNOTAVAIL: printf("Cannot assign requested address.\n");break;
	case WSAENETDOWN: printf("Network is down.\n");break;
	case WSAENETUNREACH: printf("Network is unreachable.\n");break;
	case WSAENETRESET: printf("Network dropped connection.\n");break;
	case WSAECONNABORTED: printf("Software caused connection abort.\n");break;
	case WSAECONNRESET: printf("Connection reset by peer.\n");break;
	case WSAENOBUFS: printf("No buffer space available.\n");break;
	case WSAEISCONN: printf("Socket is already connected.\n");break;
	case WSAENOTCONN: printf("Socket is not connected.\n");break;
	case WSAESHUTDOWN: printf("Cannot send after socket shutdown.\n");break;
	case WSAETIMEDOUT: printf("Connection timed out.\n");break;
	case WSAECONNREFUSED: printf("Connection refused.\n");break;
	case WSAEHOSTDOWN: printf("Host is down.\n");break;
	case WSAEHOSTUNREACH: printf("No route to host.\n");break;
	case WSAEPROCLIM: printf("Too many processes.\n");break;
	case WSASYSNOTREADY: printf("Network subsystem is unavailable.\n");break;
	case WSAVERNOTSUPPORTED: printf("Winsock dll version out of range.\n");break;
	case WSANOTINITIALISED: printf("Startup failed.\n");break;
	case WSAEDISCON: printf("Graceful shutdown in progress.\n");break;
	case WSATYPE_NOT_FOUND: printf("Class type not found.\n");break;
	case WSAHOST_NOT_FOUND: printf("Host not found.\n");break;
	case WSATRY_AGAIN: printf("Nonauthoritative host not found.\n");break;
	case WSANO_RECOVERY: printf("This is a nonrecoverable error.\n");break;
	case WSANO_DATA: printf("Valid name, no data record of requested type.\n");break;
	default: printf("Unknown error.\n");break;
	}
	return err;
}

int CEntCon::NewLinePos(const char *str, int size)
{
	if (size==0) size=(int)strlen(str);
	int i;
	for (i=0;i<size; i++)
		if (str[i]=='\n') return i;
	return -1;
}

//returns target count and a list of integers (target ids) into provided buffer
int CEntCon::List(int *pIBuf, int pLen, const char *pType, const bool pRefresh)
{
	int c=0;
	if ((!pRefresh)&&(targetIds))
	{
		int m=numTargets>pLen?pLen:numTargets;
		for (int i=0; i<m; i++)
			pIBuf[i]=targetIds[i];
		return numTargets;
	}
	char buf[2048];
	sprintf(buf,"list %s\r\n",pType);
	Send(buf);
	if (Receive(buf,2048)<=(int)strlen("list:")) return 0;
	char nm[2048];
	int p=(int)strlen("list:");
	int pp=0;
	int num;
	while (p<(int)strlen(buf))
	{//fill up buffer to return, count
		while (isdigit(buf[p]))
		{
			nm[pp++]=buf[p++];
			if (p==strlen(buf)) break;
		}
		if (pp>0)
		{
			nm[pp++]=0;
			if (c<pLen)
				if (sscanf(nm,"%d",&num)==1) pIBuf[c]=num;
			c++;
			pp=0;
		}
		p++;
	}
	if (targetIds) delete []targetIds;
	targetIds=new int[c];//allocate internal buffer
	pp=0;
	c=0;
	p=(int)strlen("list:");
	while (p<(int)strlen(buf))
	{//fill up internal buffer
		while (isdigit(buf[p]))
		{
			nm[pp++]=buf[p++];
			if (p==(int)strlen(buf)) break;
		}
		if (pp>0)
		{
			nm[pp++]=0;
			if (sscanf(nm,"%d",&num)==1) targetIds[c]=num;
			c++;
			pp=0;
		}
		p++;
	}
	numTargets=c;
	return c;
}

Point3D CEntCon::GetPos(int pId)
{
	char buf[512];
	sprintf(buf,"getpos %d\r\n",pId);
	Send(buf);
	Point3D res;
	res.x=0; res.y=0; res.z=0;
	if (Receive(buf,512)<=(int)strlen("pos():")) return res;
	int idx;
	if (sscanf(buf,"pos(%d):",&idx)!=1) return res;
	if (idx!=pId) return res;
	float x,y,z;
	if (sscanf(buf,"pos(%d):%f,%f,%f",&idx,&x,&y,&z)!=4) return res;
	res.x=x; res.y=y; res.z=z;
	return res;
}

bool CEntCon::SetPos(int pId, Point3D pPos)
{
	char buf[512];
	sprintf(buf,"setpos %d %f,%f,%f\r\n",pId,pPos.x,pPos.y,pPos.z);
	Send(buf);
	if (Receive(buf,512)<=(int)strlen("setpos():")) return false;
	int idx;
	if (sscanf(buf,"setpos(%d):",&idx)!=1) return false;
	if (idx!=pId) return false;
	float x,y,z;
	if (sscanf(buf,"setpos(%d):%f,%f,%f",&idx,&x,&y,&z)!=4) return false;
	return (x==pPos.x)&&(y==pPos.y)&&(z==pPos.z);
}

bool CEntCon::MoveBy(int pId, Point3D pDiff)
{
	Point3D oldp=GetPos(pId);
	char buf[512];
	sprintf(buf,"moveby %d %f,%f,%f\r\n",pId,pDiff.x,pDiff.y,pDiff.z);
	Send(buf);
	if (Receive(buf,512)<=(int)strlen("newpos():")) return false;
	int idx;
	if (sscanf(buf,"newpos(%d):",&idx)!=1) return false;
	if (idx!=pId) return false;
	float x,y,z;
	if (sscanf(buf,"newpos(%d):%f,%f,%f",&idx,&x,&y,&z)!=4) return false;
	return (x==oldp.x+pDiff.x)&&(y==oldp.y+pDiff.y)&&(z==oldp.z+pDiff.z);
}
