/*****************************************************************************/
/*                        PTZ Camera Scheduler                               */
/*                             (C) 2009                                      */
/*                 Adrian Ilie - adyilie@cs.unc.edu                          */
/*                   Computer Science Department                             */
/*            University of North Carolina at Chapel Hill                    */
/*                 http://www.cs.unc.edu/~adyilie                            */
/*                                                                           */
/*****************************************************************************/
/*****************************************************************************/
/* Copyright The University of North Carolina at Chapel Hill.                */
/* All Rights Reserved.                                                      */
/* Permission to use, copy, and modify this software and its documentation   */
/* for educational and research purposes without fee is hereby granted,      */
/* provided that the above copyright notice and the following paragraphs     */
/* appear in all copies.                                                     */
/*                                                                           */
/* IN NO EVENT SHALL THE UNIVERSITY OF NORTH CAROLINA AT CHAPEL HILL BE      */
/* LIABLE TO ANY PARTY FOR DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR         */
/* CONSEQUENTIAL DAMAGES, INCLUDING LOST PROFITS, ARISING OUT OF THE         */
/* USE OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN IF THE UNIVERSITY        */
/* OF NORTH CAROLINA HAVE BEEN ADVISED OF THE POSSIBILITY OF SUCH            */
/* DAMAGES.                                                                  */
/*                                                                           */
/* THE UNIVERSITY OF NORTH CAROLINA SPECIFICALLY DISCLAIMS ANY               */
/* WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF      */
/* MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.  THE SOFTWARE       */
/* PROVIDED HEREUNDER IS ON AN "AS IS" BASIS, AND THE UNIVERSITY OF          */
/* NORTH CAROLINA HAS NO OBLIGATIONS TO PROVIDE MAINTENANCE, SUPPORT,        */
/* UPDATES, ENHANCEMENTS, OR MODIFICATIONS.                                  */
/*                                                                           */
/* If you have any questions regarding this agreement or the licensing of    */
/* this technology, please send them to adyilie@cs.unc.edu                   */
/*****************************************************************************/

//interface to control entities in OVVV via Mattie's EventScripts and EntCon.py
//uses blocking TCP socket calls
//Version 1.0, first public release (June 23, 2009)

#ifndef CEntCon_H
#define CEntCon_H

#include <stdio.h>
#include <winsock.h>
#include <string.h>

#include <math.h>

class Point3D
{
public:
	float x,y,z;
	Point3D(float px=0, float py=0, float pz=0){x=px;y=py;z=pz;};
	Point3D operator+(const Point3D p){return Point3D(x+p.x,y+p.y,z+p.z);};
	void operator+=(const Point3D p){x+=p.x;y+=p.y;z+=p.z;};
	Point3D operator-(const Point3D p){return Point3D(x-p.x,y-p.y,z-p.z);};
	void operator-=(const Point3D p){x-=p.x;y-=p.y;z-=p.z;};
	float operator*(const Point3D p){return sqrt(p.x*x+p.y*y+p.z*z);};
	bool operator==(const Point3D p){return (x==p.x)&&(y==p.y)&&(z==p.z);};
	bool operator!=(const Point3D p){return (x!=p.x)||(y!=p.y)||(z!=p.z);};
	void operator=(const Point3D p){x=p.x;y=p.y;z=p.z;};
};

class CEntCon
{
public:
	CEntCon(const char *pHost="localhost", int pPort=8081);
	~CEntCon();
	int Connect();
	void Disconnect();
	int Send(const char *pBuf, int pSize=0);
	int Receive(char *pBuf, int pBufSize, int pTimeOut = 10000);
	int List(int *pIBuf, int pLen, const char *pType="npc_citizen", const bool pRefresh=true);
	Point3D GetPos(int pId);
	bool SetPos(int pId, Point3D pPos);
	bool MoveBy(int pId, Point3D pDiff);
private:
	int *targetIds,numTargets;
	static bool winSockStarted;
	char host[512];
	int port;
	sockaddr_in addr;
	SOCKET socketfd;
	bool closed;

	void StartWINSock();
	bool Probe(int timeout = 10000);
	int Error();
	int NewLinePos(const char *str, int size=0);
};

#endif