################################################################################
# EntCon: an EventScripts Python plugin for entity control in HL2
# http://www.cs.unc.edu/~adyilie/Research/CameraScheduler/EntCon/
# Part of "PTZ camera Scheduler" (C) 2009
# Adrian Ilie - adyilie@cs.unc.edu
# Computer Science Department
# University of North Carolina at Chapel Hill
# http://www.cs.unc.edu/~adyilie
# Version 1.0, first public release (June 23, 2009)
################################################################################

import sys
import traceback
import asynchat
import asyncore
import string
import socket
import es
import gamethread
import keyvalues
import playerlib

import psyco
psyco.full()

running=False
host = ''
port = 8081
server = 0
user_id = 0
# tries to get user id; if successful, entity 0 will manipulate the player

################################################################################
# EventScripts events
################################################################################

info = es.AddonInfo()
info.name = "Entity Controller"
info.version = "1.1-alpha"
info.author = "Adrian Ilie"
info.url = "http://www.adyilie.com"
info.basename = "entcon"
info.description = "Lets you control entities from outside application"

def MyFunc():
    asyncore.poll()
    global running
    if running:
        gamethread.delayed(0.1, MyFunc)
    else:
        mylog("MyFunc no longer queued")

def load():
    global running
    running = True
    mylog("loading")
    es.set("ccb_version",info.version)
    es.makepublic("ccb_version")
    mylog("Creating server socket")
    global host
    global port
    global server
    server = EntServer(host, port)
    mylog("Queueing MyFunc")
    gamethread.delayed(1, MyFunc)
    mylog("Loaded")    

def unload():
    global running
    running = False
    mylog("Unloading")
    asyncore.close_all()
    mylog("Unloaded")    

def mylog(msg):
    es.log("[EntCon] %s" % msg)
    es.msg("[EntCon]"+msg)

        
   
################################################################################
# Non-blocking multi-threaded socket server 
################################################################################

class EntServer (asyncore.dispatcher):
    def __init__ (self, host='',  port=8081):
        self.port = port
        asyncore.dispatcher.__init__ (self)
        self.create_socket (socket.AF_INET, socket.SOCK_STREAM)
        self.bind((host, port))
        self.listen(5)

    def handle_accept (self):
        conn, addr = self.accept()
        mylog ('Incoming connection from %s:%d' % (addr[0], addr[1]))
        EntChannel (conn, addr)

class EntChannel(asynchat.async_chat):
    def __init__ (self, conn, addr):
        asynchat.async_chat.__init__ (self, conn)
        self.set_terminator ('\r\n')
        self.in_buffer = ''
        global user_id
        user_id = es.getuserid()

    def handle_connect (self):
        self.send('\377\375\"')#TelNet magic string

    def collect_incoming_data (self, data):
        self.in_buffer = self.in_buffer + data

    def found_terminator (self):
        mylog("Received "+self.in_buffer)
        cmd = ''
        ent = ''
        vstr='0.0,0.0,0.0'
        fvec = [0.0, 0.0, 0.0]
        svec=['0.0', '0.0', '0.0']
        lines = string.split(self.in_buffer,' ')
        leng = len(lines)
        if leng>0:
            cmd = lines[0]
            if leng>1:
                ent = lines[1]
            if leng>2:
                vstr = lines[2]
                svec = vstr.split(',')
                fvec = (float(svec[0]),float(svec[1]),float(svec[2]))
            global user_id
            if cmd=="list":  # list entities
                c=0
                res = ''
                if user_id != 0:
                    res = "0";
                    c=1
                for i, e in  es.createentitylist(ent).items():
                    if c>0:
                        res = ',' + res
                    res = "%d"%i + res
                    c += 1
                res = 'list:'+res
                self.reply(res)
            elif cmd=="getpos":  # get position of entity
                if ent=='0':
                    pos = es.getplayerprop(user_id, "CBaseEntity.m_vecOrigin")
                else:
                    pos = es.getindexprop(ent, "CBaseEntity.m_vecOrigin")
                self.reply("pos("+ent+"):"+pos)
            elif cmd=="setpos":  # set position of entity
                if ent=='0':
                    es.setplayerprop(user_id, "CBaseEntity.m_vecOrigin", vstr)
                    pos = es.getplayerprop(user_id, "CBaseEntity.m_vecOrigin")
                else:
                    es.setindexprop(ent, "CBaseEntity.m_vecOrigin", vstr)
                    pos = es.getindexprop(ent, "CBaseEntity.m_vecOrigin")
                self.reply("setpos("+ent+"):"+pos)
            elif cmd=="getang":  # get angle of entity
                if ent=='0':
                    ang = es.getplayerprop(user_id,"CBaseEntity.m_angRotation")
                else:
                    ang = es.getindexprop(ent, "CBaseEntity.m_angRotation")
                self.reply("ang("+ent+"):"+ang)
            elif cmd=="setang":  # set angle of entity
                if ent=='0':#doesn't work for player :(
                    es.setplayerprop(user_id, "CBaseEntity.m_angRotation", vstr)
                    ang = es.getplayerprop(user_id,"CBaseEntity.m_angRotation")
                else:
                    es.setindexprop(ent, "CBaseEntity.m_angRotation", vstr)
                    ang = es.getindexprop(ent, "CBaseEntity.m_angRotation")
                self.reply("setang("+ent+"):"+ang)
            elif cmd=="moveby":  # move entity by a delta vector
                if ent=='0':
                    pos = es.getplayerprop(user_id, "CBaseEntity.m_vecOrigin")
                else:
                    pos = es.getindexprop(ent, "CBaseEntity.m_vecOrigin")
                spos = pos.split(',')
                fpos = (float(spos[0]),float(spos[1]),float(spos[2]))
                nfpos = (fpos[0]+fvec[0],fpos[1]+fvec[1],fpos[2]+fvec[2])
                nspos = "%f,%f,%f"%(nfpos[0], nfpos[1], nfpos[2]) 
                if ent=='0':
                    es.setplayerprop(user_id, "CBaseEntity.m_vecOrigin", nspos)
                    pos = es.getplayerprop(user_id, "CBaseEntity.m_vecOrigin")
                else:
                    es.setindexprop(ent, "CBaseEntity.m_vecOrigin", nspos)
                    pos = es.getindexprop(ent, "CBaseEntity.m_vecOrigin")
                self.reply("newpos("+ent+"):"+pos)
            elif cmd=="echo":  # echo sent command
                self.reply(ent)
            elif cmd=="quit":
                global running
                running = False
            else:
                self.reply("badcmd:"+cmd)
            self.in_buffer = ''
        else:
            mylog("0-length request")
            self.close()

    def handle_close (self):
        mylog("Channel closing")
        self.close()
        
    def reply(self, rep):
        self.send(rep+'\r\n')
        mylog("Sent "+rep)
